;;; linkxall.el --- Provides hyperlinking with LibreOffice LinkXall library
;; Copyright (C) 2022-2024 David McCracken
;; Author: David McCracken
;; Keywords: lisp LibreOffice LinkXall
;; Version: 0.1.0

;; This program is free software; you can redistribute it and/or modify it 
;; under the terms of the GNU General Public License version 3 as published by 
;; the Free Software Foundation

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;; Updated:  2024.04.10.11.16
;; Purposes:  Emacs LinkXall library. This works with the LibreOffice Basic 
;; LinkXall library to implement plain text LinkXall-compliant hyperlinks in 
;; both Windows and Linux (with X). 

;; Extended program comments are in LO doc files referenced here by links.
;; *LXA-ODF*|LXADOC|LxaUserGuide.odt*^1Emacs* General description in User Guide.
;; *LXA-ODF*|LXADOC|LxaUserGuide.odt*^2EmacsUserInterface* Menu Help.
;; *LXA-ODF*|LXADOC|LxaDesign.odt*^1Emacs* Program design

;; Place this in Emacs\share\emacs\site-lisp and auto-load it with 
;; (load-library "linkxall.el" ) in .emacs. Also, all of the code is wrapped in
;; a progn, which can be interactively evaluated to test program changes.
;; Icons for the toolbar are located in ~/icons. In Windows, HOME evironment
;; variable must be defined, as e.g. C:\Home. The icons are lxa-backlink.xpm,
;; lxa-fwdlink.xpm, lxa-pin.xpm, lxa-prev.xpm, and lxa-next.xpm.

;; When this loads, it creates LXA menu and toolbar and assigns keyboard and 
;; mouse shortcuts. There are five major UI actions, open/follow, Back, Forward,
;; Previous, and Next. Open/follow is the only context-sensitive action. It 
;; opens the menu at the caret on M-l (downcase-word is rebound to M-L) or 
;; pointer on C-S-mouse-1 unless the text is a (syntactically) valid LinkXall 
;; link, in which case the link is "followed" by passing it to LO LinkXall. 
;; Back and Forward peruse the jump list, which is maintained by LinkXall. 
;; Keyboard shortcuts M-S-left and M-S-right are bound to Back and Forward, M-S-
;; up and M-S-down to Previous and Next. Previous and Next search for the 
;; previous and next reference, which is either a bookmark or link. 

;; By default this asks the LinkXall (Libre BASIC) library to open all link 
;; types. However, it can follow ODT, BSR, and WEB types itself and doing so may 
;; afford better performance in some circumstances. This "mixed" mode (BAS,TXT, 
;; and SYS require the library) is enabled by replacing the t with nil in
;; (setq lxa-min-follow t).

;*******************************************************************************

(progn ; Wrapper to invoke all definitions during development.

; ============================== UTILITIES =====================================

;------------------------------- #move-mouse-to-point ------------------------
(defun move-mouse-to-point()
  (let ((pos (window-absolute-pixel-position)))
    (set-mouse-absolute-pixel-position
      (car pos) (cdr pos))))

; ------------------------------- #lxa-collapse -------------------------------
(defun lxa-collapse(beg)
  "Collapse selection to begin (BEG=t) or end (BEG=nil)"
  (if (xor (> (point) (mark)) beg) (set-mark (point))
    (set-window-point nil (mark))))

;------------------------------- #lxa-trim-string -------------------------------
(defun lxa-trim-string(str)
"Create new string from STR with leading and trailing space removed"
  (replace-regexp-in-string "^\s+" "" 
  (replace-regexp-in-string "\s+$" "" str)))

;----------------------------- #lxa-condense-string -----------------------------
(defun lxa-condense-string(str)
"Return nil if trimmed STR is empty else `lxa-trim-string' STR"
    (if (not (stringp str)) nil
      (let ((ns (lxa-trim-string str)))
        (if (equal ns "") nil ns))))

;----------------------------- #lxa-get-cursor-line -----------------------------
(defun lxa-get-cursor-line()
  (save-excursion
    (beginning-of-line)
    (setq beg (point))
    (end-of-line)
    (lxa-trim-string (buffer-substring-no-properties beg (point)))))

;------------------------------- #lxa-index-list -------------------------------
(defun lxa-index-list( thing reflist )
"Return the 0-based index of THING in REFLIST or -1 if not in REFLIST.\n\
Typically, THING is a string and REFLIST is a list of strings"
  (let ((listlen (length reflist)) (mlist (member thing reflist)))
    (if mlist (- listlen (length mlist)) -1 )))

(defun lxa-show-match (backward)
  "Highlight the last search match"
  (if backward ; after backward search
    (progn
      (goto-char (match-end 0))
      (set-mark (point))
      (goto-char (match-beginning 0)))
    ; else after forward search
      (goto-char (match-beginning 0))
      (set-mark (point))
      (goto-char (match-end 0))))

;------------------------------- #get-line-number -------------------------------
; Emacs version 26 defines line-number-at-pos as optionsl POS ABSOLUTE and
; version 25 as optional POS. In v26 ABSOLUTE must be t or the line number
; is based on display. v25 is automatically abs without the argument but
; complains when given two args.
(defun get-line-number() (line-number-at-pos nil t))
  (condition-case nil
    (get-line-number)
    (error (defun get-line-number() (line-number-at-pos))))

;------------------------------ #lxa-focus-window ------------------------------
(defun lxa-focus-window( uri )
  (if (> lxa-focus 0) 
    (pcase lxa-focus (1 (call-process "wmctrl" nil 1 nil "-a" uri))
      (2 (call-process "wintop" nil 1 nil uri)))))

; ===================== CORE FUNCTIONS =========================================

;----------------------------- #lxa-xcmd -----------------------------------
; External command request. Typically from OO linkxall through wemecs or lemacs.
; Return is irrelevant. 
;- cmd integer tells what to do. 
;- parm is string to search for or line (as string int) to go to.
; All arguments are quoted strings because the interface program (wemacs or 
; lemacs) doesn't know what they are when it builds the invocation command. cmd
; is required. parm and parm2 are optional. parm2 is unused but declared to 
; afford the (external) caller more freedom in formulating the command line. 
;
; cmd is bitmap where:
; b3 0 = search, 1 = go to line. If go to line then b0 = 0 means don't 
; highlight the line and 1 means do highlight it. If search then:
; b2: 0 = case-sensitive, 1 = case-insensitive. 
; b1 = start at: 0 = current, 1 = begin if forward (b0=0), end if backward (b0=1).
; b0 = direction; 0 = forward, 1 backward
; Thus:
; 0 = search forward from current case-sensitive (4 = insensitive)
; 1 = search backward from current case-sensitive (5 = insensitive)
; 2 = search forward from begin case-sensitive (6 = insenstive)
; 3 = search backward from end case sensitive (7 = insenstive)
; 8 = go to line number parm and don't highlight the line.
; 9 = go to line number parm and highlight the line.
;
; If not from current then go to end if backward or beginning if forward. If 
; from current then collapse the current selection to begin if backward or
; to end if forward to avoid repeatly finding the same instance.
; The search function is invoked with limit=nil and noerror=t, which is very
; important because with default noerror=nil on search failure this function
; is aborted and an error code returned to the calling program, which will not
; be able to distinguish it from emacsclient failure.
; *LXA-ODF*|LXADOC|LxaDesign.odt*^3lxa-xcmd*
; *LXA-BAS*LinkXall>LinkXall*#LXAopenEmacs*
; .............................................................................
(defun lxa-xcmd(cmd &optional parm parm2)
"linkxall external command request"
  (interactive)
  (let* 
    ((cnum (string-to-number cmd))
      (b0 (= (logand cnum 1) 1)) ; b0 = t if cmd b0 = 1
      (b1 (= (logand cnum 2) 2)) ; b1 = t if cmd b1 = 1
      (case-fold-search (= (logand cnum 4) 4)))
    (when parm ; parm is required in all of the following  cases. 
      (if (< cnum 8) ; Search for parm string
        (progn 
          (if b1 (if b0 (end-of-buffer) (beginning-of-buffer)) (lxa-collapse b0))
          (if (funcall (if b0 'search-backward 'search-forward) parm nil t)
            (progn (lxa-show-match nil) (setq lxa-search parm))
          ; else not found. Collapse selection to avoid its extension and display error. 
            (set-window-point nil (mark))(message-box (concat "Unable to find " parm))))
      ; else (cnum >= 8) goto to parm line number
        (goto-line (string-to-number parm))
        (when b0 (set-mark (point)) (end-of-line))) ; eol), when b0, if
      ))t) ; when parm, let, defun

;------------------------- #lxa-back-fwd ----------------------------------
; *LXA-ODF*|LXADOC|LxaDesign.odt*^3LXAxrev*
(defun lxa-back-fwd(bf)
  "Shared core of lxa-link-back and lxa-link-forward, used to peruse the \
LXA link trail. BF is 'B' to go backward and 'F' to go forward. These specific \
values are required for coordinating with other programs." 
  (let ((linkproc))
    (setq linkproc
    (start-process "openlinkproc" nil 
      (if (getenv "ODAPP") (substitute-in-file-name "$ODAPP") "soffice")
      (concat "macro:///LinkXall.LinkXall.LXAxrev(\"" bf  "\")") ))
    (set-process-query-on-exit-flag linkproc nil)))

;-------------------------------- #lxa-link-back --------------------------------
(defun lxa-link-back()
"Go back to previous link point"
  (interactive)
  (lxa-back-fwd "B"))

;------------------------------ #lxa-link-forward ------------------------------
(defun lxa-link-forward()
"Go to next forward link point"
  (interactive)
  (lxa-back-fwd "F"))

;----------------------------- #lxa-set-pinpoint ----------------------------
(defun lxa-set-pinpoint()
"Record point for revisit in jump list"
  (interactive)
  (let ((linkproc))
    (setq linkproc
      (start-process "openlinkproc" nil 
        (if (getenv "ODAPP") (substitute-in-file-name "$ODAPP") "soffice")
        (concat "macro:///LinkXall.LinkXall.LXAxlink(LXA-PIN;"
          buffer-file-name ";" (number-to-string (get-line-number)) "*2)")))
; *LXA-ODF*|LXADOC|LxaDesign.odt*^3LXAxlink*
; *LXA-BAS*LinkXall>LinkXall*#LXAxlink
      (set-process-query-on-exit-flag linkproc nil)))

; ---------------------- #lxa-prev-next --------------------------
(setq lxa-search nil)

(defun lxa-prev-next(pn)
  "Shared core of lxa-prev and lxa-next functions to search for the most \
recent LXA reference. PN is t for previous (search backward) and nil for \
next (search forward)"
  (if (equal lxa-search nil) (message-box "There is no reference")
  (if (use-region-p) (goto-char (if pn (region-beginning) (region-end))))
  (when (funcall (if pn 'search-backward 'search-forward) lxa-search nil nil)
    (lxa-show-match t) (message lxa-search))))

;---------------------------------- #lxa-prev ----------------------------------
(defun lxa-prev()
"Search backward for the lxa reference"
  (interactive)
  (lxa-prev-next t))

;---------------------------------- #lxa-next ----------------------------------
(defun lxa-next()
"Search forward for the lxa reference"
  (interactive)
  (lxa-prev-next nil))

;---------------------------------- #lxa-help-ui ----------------------------------
(defun lxa-help-ui()
"User interface help"
  (lxa-open-link "*LXA-ODF*|LXADOC|LxaUserGuide.odt*^2EmacsUserInterface*" t ))

;---------------------------------- #lxa-help-bind ----------------------------------
(defun lxa-help-bind()
"Bindings help"
  (lxa-open-link "*LXA-ODF*|LXADOC|LxaUserGuide.odt*^2Bindings*" t ))

;------------------------------ #lxa-open-err ------------------------------
(defun lxa-open-err()
  (message-box "what")) ; "There is no LXA link in \"%s\"" (lxa-get-cursor-line)))

;--------------------- #lxa-open-link -----------------------------------
; Open (follow) an LXA-type link. 
; Return 0 if the text doesn't contain an LXA link type, 1 if it does but not a 
; completely valid link (syntax error or missing piece), 2 if the link URL 
; includes an undefined environment variable, 3 if the link is completely valid 
; but the target can't be opened, and 4 on complete success. 
; Optional Arguments:
;- txt is the text to be evaluated. If not provided then the line containing the
;  cursor is evaluated.
;- rpt t tells to report to the user all errors processing the text as a link.
;  Missing or nil tells to report only errors trying to follow a syntactically 
;  valid link.
; *LXA-ODF*|LXADOC|LxaDesign.odt*^3lxa-open-link*
; ...........................................................................
(defun lxa-open-link(&optional txt rpt) 
"Open a link to a LibreOffice document or macro library, another text file,\n\
local HTML file, or web URL. This requires LibreOffice with LinkXall library.\n\
Optional TXT is a LinkXall text link. If missing then the entire line\n\
containing the cursor is processed. Optional RPT if t tells to report\n\
all errors to the user. If missing or nil then only interpretation errors\n\
are reported."
  (interactive)
  (unless txt (setq txt (lxa-get-cursor-line)))
  (catch 'fatal
    (setq lxa-err 0)
    (let* ( (oswin (equal system-type 'windows-nt)) 
            (oslin (not oswin))
            (linktype -1) 
            (args (split-string txt "*" nil))
            beg name uri bm linkproc wcnt outbuf kbqf here nopro )
      (while 
        (progn
          (setq name (pop args))
          (and name
            (< (setq linktype 
                 (lxa-index-list name 
                   (list "LXA-BAS" "LXA-ODF" "LXA-TXT" "LXA-BSR" "LXA-SYS" "LXA-WEB")
                   ))0))))
      (when (< linktype 0) 
        (when rpt (message-box "No LXA link type in \"%s\"" txt))
        (throw 'fatal 0))
      (setq uri (lxa-condense-string (pop args)))
      (when (not uri)
        (when rpt (message-box "LXA link missing URI in \"%s\"" txt))
        (throw 'fatal 1))
      (setq bm (lxa-condense-string (pop args)))

      (setq here 
        (concat buffer-file-name ";" 
          (number-to-string (get-line-number))))

; ---------------- minimal form -------------------------------------------------
; If configured to use only min form or if linktype is only supported by min
; (0/BAS, 2/TXT, 4/SYS).
      (if (or lxa-min-follow (> (lxa-index-list linktype (list 0 2 4)) -1))
        (progn
;(message (concat "macro:///LinkXall.LinkXall.LXAxlink(" name "*" uri "*" bm ";"  here "*2)")) ; ))))))
          (setq linkproc 
            (start-process "openlinkproc" nil 
              (if (getenv "ODAPP") (substitute-in-file-name "$ODAPP") "soffice")
              (concat "macro:///LinkXall.LinkXall.LXAxlink(" 
                name "*" uri "*" bm ";"  here "*2)")))
          (set-process-query-on-exit-flag linkproc nil)
          (if lxa-min-focus 
            (lxa-focus-window (if (= linktype 0) "LibreOffice Basic" 
                                (file-name-nondirectory uri))))
          ) ; min form progn

;--------------- else long form  ---------------------
        (setq nopro (not (string-match ":///" uri)))
        (if nopro ; Expand/test uri unless (WEB) protocol.
          (progn
            (setq uri 
              (replace-regexp-in-string "|[^|]+|" 
; If the environment variable is not in the dictionary, lxa-getenvar throws
; 'fatal 3.
                (lambda(var) (lxa-getenvar (substring var 1 -1)))
                uri t)) ; Note t fixedcase optional argument to 
; replace-regexp-in-string. Otherwise it uses the case of the varname.
            (setq uri ; If relative convert to absolute.
              (expand-file-name uri (file-name-directory buffer-file-name)))
            (if (not (file-readable-p uri))
              (progn (message-box (concat uri " does not exist.")) 
                (throw 'fatal 3)))))
        
; If WEB or BSR when DEFBROWSER is not defined then use browse-url. This 
; function always returns nil. Failure will be silent unless the system reports 
; it. LinkXall supports both with and w/o protocol but browse-url fails in Linux
; without file:/// so this is prefixed to uri if it has no protocol. This doesn't
; hurt in Windows.
        (if (or (= linktype 5) ; WEB
              (and (= linktype 3) ; BSR
                (null lxa-browser)))
          (browse-url (concat (if nopro "file:///") uri 
                        (if (and bm oslin)(concat "#" bm))))
;(message-box (concat (if nopro "file:///") uri (if (and bm oslin)(concat "#" bm))))
; Else LXA-ODF or LXA-BSR with DEFBROWSER defined.
          (if (= linktype 1) ; ........... LXA-ODF ....................
            (progn
;(message (concat uri " macro:///LinkXall.LinkXall.LXAxlink(" bm ";" here "*2)")) ; )))))))
              (setq linkproc 
                (start-process "openlinkproc" nil 
                  (if (getenv "ODAPP") 
                    (substitute-in-file-name "$ODAPP") "soffice")
                  uri
                  (concat "macro:///LinkXall.LinkXall.LXAxlink(" bm ";" here "*2)")))
;(if (equal bm nil) ";" bm) "," here "*2)")))
              (set-process-query-on-exit-flag linkproc nil)
; lxa-wmctrl is t if wmctrl is installed, which also means this is Linux, in 
; which case, if soffice is already open, the document window won't be brought 
; forward. We can do that with wmctrl. -a tells to make the specified window 
; top-most. 
;(if lxa-wmctrl (call-process "wmctrl" nil 1 nil "-a" 
;                (file-name-nondirectory uri)))
              ) ; ODT progn
            
; Else ............... LXA-BSR with bm in Windows ..................
            (setq uri (concat "file:///" uri))
            (setq outbuf "*openlinkbuf*")
            (setq wcnt (length (window-list)))
            (async-shell-command 
              (concat 
                (shell-quote-argument lxa-browser)
                " " uri "#" bm) outbuf outbuf)
            (if (bufferp (get-buffer outbuf))
              (progn
                (sleep-for 0.5)
                (setq kbqf kill-buffer-query-functions)
                (setq kill-buffer-query-functions nil)
                (kill-buffer outbuf)
                (setq kill-buffer-query-functions kbqf) ; This is supposed to
; suppress "active processes" query if closing Emacs with the spawned app still
; open but it doesn't work.
                (if (equal wcnt 1) (delete-other-windows))))
            ) ; ODT and BSR
          ) ; If WEB else ODT/BSR
        (lxa-focus-window uri) ; Unconditionally in long form
        ) ; End of long form clause and follow-form if.
      4 ) ; No problems.
    ) ; catch
  ) ; lxa-open-link

; ------------------------ #lxa-lxa --------------------------------------
; If there is a valid LXA link at the point then open it. Otherwise open the 
; LinkXall menu at the point. To allow the menu to be opened in a line 
; containing a valid link, if a region has been selected then unconditionally 
; open the menu. The link may still be followed using the menu's Follow link.
; *LXA-ODF*|LXADOC|LxaDesign.odt*^3lxa-lxa*
; .....................................................................
(defun lxa-lxa()
  "Process LXA request at the cursor. If the text contains an LXA link \
then open it. Otherwise, or if a region has been selected, open the LinkXall \
menu at the cursor."
  (interactive)
  (when (or (use-region-p)(< (lxa-open-link nil nil) 4))
    (move-mouse-to-point)
    (let ((sel (x-popup-menu t lxa-menu)))
      (pcase (car (split-string (format "%s" (car sel))))
        ("Back" (lxa-link-back))
        ("Forward" (lxas-link-forward))
        ("Set" (lxa-set-pinpoint))
        ("Previous" (lxa-prev))
        ("Next" (lxa-next))
        ("Follow" (lxa-open-link nil t))
        ("Start" (server-start nil t))
        ("Help" 
          (pcase (car (split-string (format "%s" (car (cdr sel)))))
            ("User" (lxa-help-ui))
            ("Bindings" (lxa-help-bind))))))))

; ------------------------------ #lxa-menu ---------------------------------
  (easy-menu-define lxa-menu nil "lxa-menu"
    '(lxa-menu
       ["Back link\tC-, (<)" (lxa-link-back)]
       ["Forward link\tC-. (>)" (lxa-link-forward)]
       ["Set pinpoint" (lxa-set-pinpoint)]
       ["Previous reference\tM-S-up" (lxa-prev)]
       ["Next reference\tM-S-down" (lxa-next)]
       ["Follow link\tM-l (L) or C-S-leftClick" (lxa-open-link nil t)] ;(lxa-lxa t)]
       ["Start server" (server-start nil t)]
       ("Help" 
         ["User interface" (lxa-help-ui)]
         ["Bindings" (lxa-help-bind)])))
; At right end (between Tools and Emacs-Lisp)
  (define-key-after (lookup-key global-map [menu-bar])
    [lxa-menu]
    (cons "LinkXall" lxa-menu) t) ; t, 'file, 'edit, etc. or empty.
; At left end
;  (define-key (lookup-key global-map [menu-bar])
;    [lxa-menu]
;    (cons "LinkXall" lxa-menu))

; ========================= ENV DICTIONARY =================================
; These functions are needed only if Emacs follows one or more LXA link types 
; itself instead of passing the text to LO LinkXall to do everything.
; *LXA-ODF*|LXADOC|LxaDesign.odt*^2EmacsEnvironmentDictionary*

;--------------------------- #lxa-getenvar -----------------------------------
(defun lxa-getenvar(str &optional silent)
"Get the lxa private or public environment variable value"
  (let (( val (if lxaenvars (cdr (assoc str lxaenvars)))))
    (if (not val) (setq val (subst-char-in-string ?\\ ?/ (getenv str))))
    (if (or (not val) (string= val ""))
      (if silent nil
        (message-box (concat str " environment variable is undefined"))
        (throw 'fatal 2)) 
      val)))

;--------------------------- #lxa-read-env -----------------------------------
(defun lxa-read-env()
"Make lxaenvars alist from the envar.txt file in the $LINKXALL_ENV directory.\n\
If the standard environment variable LINKXALL_ENV is undefined or \n\
$LINKXALL_ENV/envar.txt does not exist, lxaenvars is nulled and nil is returned.\n\
Otherwise, t is returned. This automatically executes when the library loads but\n\
it may also be invoked interactively to update the dictionary after editing the file."
  (interactive)
; Verify dictionary creation by C-h V lxaenvars, which shows the list of cons cells.
  (setq lxaenvars nil)
  (let ((fspec (getenv "LINKXALL_ENV")))
    (if (not fspec) nil
      (setq fspec (concat fspec "/envar.txt"))
      (if (not (file-readable-p fspec)) nil
        (with-temp-buffer
          (insert-file-contents fspec)
          (goto-char (point-min))
          (while (not (eobp)) 
           (re-search-forward "^ *\\([^=]+\\)=\\(.+\\)$" nil t)
; The arguments to re-search-forward are RE limit noerror. nil limit means end 
; of buffer. This is just a place-holder enabling noerror t, which means to 
; return nil but not signal an error if the match fails. This is critical for 
; blank lines, which don't mean anything bad or good.
            (if (< (aref (match-string 1) 0) ?A) nil
              (setq lxaenvars 
                (append lxaenvars 
                  `(( ,(match-string 1) . 
                      ,(subst-char-in-string ?\\ ?/ (match-string 2))))))
              (forward-line 1))))t)))) ; lxa-read-env

; ============================= INITIALIZATION  ==============================
;--------------------- #lxa-mouse-open -------------------------
(defun lxa-mouse-open()
"Move cursor to mouse position and invoke `lxa-lxa'"
  (interactive) ; Required for binding to mouse button
  (mouse-set-point last-input-event)
  (redisplay t)
  (lxa-lxa)
) ; lxa-mouse-open

;---------------- #hot keys ---------------------------------------
(global-set-key "\M-L" 'downcase-word) ; Rebind to free \M-l.
(global-set-key "\M-l" 'lxa-lxa)   
(global-set-key [C-S-mouse-1]#'lxa-mouse-open)

(global-unset-key (kbd "C-,")) 
(global-set-key (kbd "C-,") 'lxa-link-back) 

(global-unset-key (kbd "C-.")) 
(global-set-key (kbd "C-.") 'lxa-link-forward)

(global-set-key [M-S-up]#'lxa-prev)
(global-set-key [M-S-down]#'lxa-next)

; ------------------------ #toolbar --------------------------------
(define-key-after (default-value 'tool-bar-map)
  [separator-4] menu-bar-separator) 

; Linux-Emacs v25 includes pseudo-key name in the toolbar unless it is too
; long, which is the only way to stop this unwanted behavior.

(tool-bar-add-item "~/icons/lxa-backlink"  
  'lxa-link-back         ; function
  'lxa-external-link-back ; pseudo-key name 
  :help "LXA external link back")

(tool-bar-add-item "~/icons/lxa-fwdlink"  
  'lxa-link-forward  
  'lxa-external-link-forward
  :help "LXA external link forward")

(tool-bar-add-item "~/icons/lxa-pin"  
  'lxa-set-pinpoint  
  'lxa-external-set-pinpoint
  :help "LXA set pinpoint")

(tool-bar-add-item "~/icons/lxa-prev"  
  'lxa-prev  
  'lxa-prev-reference 
  :help "LXA previous reference")

(tool-bar-add-item "~/icons/lxa-next"  
  'lxa-next  
  'lxa-next-reference 
  :help "LXA next reference")

; ------------------------- #Environment ----------------------------------
; Environment dictionary and focus are related but independent. If Emacs follows
; any link type itself then it must be able to focus the target but even if it 
; asks LO LinkXall to handle all links, window focus may still be more reliable
; if done here instead of (or in addition to) in LO.
;- lxa-min-follow t means use min follow form for all. If nil then mixed, with 
; min form used for BAS, TXT, and SYS and long form for the others.
;- lxa-min-focus t means that min form tries to focus the target window. This
; applies regardless of whether all min or mixed. Focus is always done in long
; form.
; ........................................................................
  (setq lxa-min-follow t) ; t = use only min form, nil = mixed.
  (unless lxa-min-follow ; Only mixed needs this.
    (lxa-read-env)
    (setq lxa-browser (lxa-getenvar "DEFBROWSER" t)))
  ;(if lxa-browser (message-box lxa-browser))
  (setq lxa-min-focus nil) ; t = focus target window in min form. This is done
; only because LinkXall does not do it or is unsuccessful in some cases.
  (setq lxa-focus (if (executable-find "wmctrl") 1 
                      (if (executable-find "wintop") 2 0)))
t ) ; End of progn wrapper. Place cursor at close paren and type C-x C-e.
;; =========================================================================
(provide 'linkxall)


;=========================================================================
; LinkXall uses four public environment variables related to Emacs. 
; EMACSEXEPATH is the path to emacs, emacsclient, and (in Windows) runemacs.
; This must include a trailing delimiter, e.g. "C:/something/emacs/bin/"

; ODAPP is soffice filespec, e.g. "C:/Program Files/LibreOffice/program/soffice"

; LINKXALL_ENV tells the location of LinkXall's private environment file 
; envar.txt, e.g. "C:/Home/LinkXall/" or "/home/user/linkxall". This needs to
; be defined only if linkxall.el is configured for mixed form and expands
; environment variables itself. 

; DEFBROWSER is the preferred browser filespec, e.g. 
; "C:/Program Files/Mozilla Firefox/firefox.exe" or "/usr/bin/firefox". 
; If undefined then the system default browser will be used to follow LXA-BSR 
; links (it is always used for LXA-WEB links). DEFBROWSER may alternatively be 
; defined in envar.txt. DEFBROWSER mainly exists to address Windows lack of 
; support for standard bookmark (partial-URL) syntax "#bookmark". It also 
; enables using  different browswers for local files vs. Web. This is needed 
; here only if configured for mixed form. 

; EMACSEXEPATH and ODAPP exist mainly exist for Windows, which doesn't 
; automatically put these programs in the path. However, they are also accepted
; in Linux. If undefined (in either OS) the programs are invoked just by name
; and must be in the path. EMACSEXEPATH is used by wemacs.exe, lemacs, and 
; LinkXall library. Only linkxall.el uses ODAPP.

; LINKXALL_ENV is the only variable that exists for a reason other than to 
; correct a Windows problem. It is typically the only one of these variables be 
; be defined in Linux and then only if linkxall is configured for mixed form.
;
; Use the following temporaries for testing. After changing LINKXALL-ENV, eval 
; the all-encompassing progn to build the dictionary.

;(setenv "ODAPP" "C:/Program Files/LibreOffice/program/soffice") ; Windows
;(setenv "DEFBROWSER" "C:/Program Files/Mozilla Firefox/firefox.exe") ; Windows
;(setenv "LINKXALL_ENV" "C:/Home/LinkXall") ; Windows
;(setenv "LINKXALL_ENV" "~/linkxall") ; Linux

; To make these semi-permanent make them user variables in System Properties > 
; Advanced > Environment variables. They will appear in any subsequently opened 
; windows. Linux is more complicated. Ubuntu recommends putting them in a 
; startup script, e.g. /etc/profile.d/linkxall.sh containing export 
; LINKXALL_ENV=~/linkxall. These take effect on reboot but full system reboot 
; is not required. sudo systemctl restart display-manager works in most systems 
; but the x-manager may need to be explicitly named, e.g. sudo systemctl 
; restart lightdm. Get the name by cat /etc/X11/default-display-manager 

; ====================== TEST CASES =======================================
; Min form suppports all LXA types BAS, ODT, TXT, BSR, SYS, and WEB. Long form 
; supports only ODT, BSR, and WEB. These are tested for both min and long. BAS,
; TXT, and SYS can only be tested for min form, which is automatically selected 
; regardless of lxa-min-follow value.

; Syntax and missing file error tests can be OS-agnostic. Tests that are 
; expected to succeed can use environment or relative addressing. Absolute must 
; be OS-specific.

; ..................... Syntax Errors ..................................
; Syntax errors are reported here for all types if open link is invoked from 
; the menu. With KB/mouse shortcuts they just cause the context menu to open.
; *LXA-JUNK*someFile** Windows and Linux. No link here.
; *LXA-ODF*** Windows and Linux. No uri.

; ..................... Inaccessible Target .................................
; Missing file is reported here (long form) or in LinkXall (min form) except
; for LXA-WEB with protocol (even file:///) which may fail silently in Linux.
; *LXA-ODF*badFile.odt** Windows and Linx. min and long form.
; *LXA-BSR*badFile.odt** Windows and Linx. min and long form. DEFBROWSER is
;   irrelevant.
;   ShellExecute error (in echo area). min form displays system error.
; *LXA-BAS*LinkXall>NoLib*WhoCares* Only min form. LinkXall error message.
; *LXA-TXT*noFile*NotMe* Only min form. LinkXall error message.
; *LXA-SYS*noFile*NotMe* Only min form. LinkXall error message.
; *LXA-WEB*badFile.odt** Windows and Linux. min and long form.
; *LXA-WEB*file:///badFile.htm** Windows and Linux. In Windows long form reports
;   in echo line, min in system error box. Fails silently with either form in
;   Linux.

; ...................... Address Forms ...................................
; These are all done with ODT because that can be tested for both min and
; long forms and the canonical test of all forms with all link types is
; too much since this is not automated.
; *LXA-ODF*test/bm3.odt*bm2* Relative
; *LXA-ODF*|SWDEV|Emacs/test/bm3.odt*bm2* Private environment.
; *LXA-ODF*|HOMEPATH|/linkxall.odt*bm3* Windows public environment.
; *LXA-ODF*|SSH_AUTH_SOCK|/linkxall.odt*bm3* Linux public environment
; *LXA-ODF*c:/SwDev/Emacs/test/bm3.odt*bm2* Absolute Windows
; *LXA-ODF*/home/d/work/SwDev/Emacs/test/bm3.odt*bm2* Absolute Linux
; *LXA-ODF*/SwDev/Emacs/test/bm3.odt*bm2* Root Windows

; ....................... Min form only ..................................
; BAS, TXT, and SYS can only be processing by min form. The only option here
; is whether we try to focus the target window. Since all but simple syntax
; verification is done in LinkXall, it would be reasonable for LinkXall to
; focus the window but it may not be able to reliably do this when acting as
; hyperlink server.
; *LXA-BAS*LinkXall>LinkXall*#LXAopenLink
; *LXA-TXT*test/linkxall.el*Min form only* Relative addressing.
; *LXA-TXT*|SWDEV|Emacs/test/linkxall.el*Min form only* Env addressing.
; *LXA-SYS*test/getch.py** Relative addressing
; *LXA-SYS*|SWDEV|Emacs/test/getch.py** Env addressing

; .......................... ODT ..........................................
; *LXA-ODF*|SWDEV|/Libre/LinkXall/linkXall.odt*bm3
; *LXA-ODF*|SWDEV|Libre/LinkXall/bm 4 .odt*bm8
; *LXA-ODF*../Libre/LinkXall/linkxall.odt*bm3*comment
; *LXA-ODF*|HOMEPATH|/work/test/bm3.odt*bm2* Windows public environment.
; *LXA-ODF*C:/home/work/test/bm3.odt*Bm30*Windows absolute
; *LXA-ODF*C:\home\work\test\bm3.odt*Bm30*Windows absolute
; *LXA-ODF*/home/d/work/test/bm3.odt*Bm30*Linux absolute
; Relative with bm  w/o comment *LXA-ODF*test/bm3.odt*firstBm*
; Relative with bm w/o comment *LXA-ODF*test/bm3.odt*firstBm
; *LXA-ODF*test/bm3.odt*firstBm *Relative with commebNt after bm
; *LXA-ODF*test/bm3.odt*   firstBm   * bm in space
; *LXA-ODF*test/bm3.odt** Relative with comment after null bm
; *LXA-ODF*test/bm3.odt* * Comment after blank bm
; *LXA-ODF*~/work/test/bm3.odt*Bm30* Fails. ~ is ordinary.
; *LXA-ODF*|HOME|/work/test/bm3.odt*bm2* Private or public  environment variable
; *LXA-ODF*|LIBRARY|/bm1test.odt*bm1* Private environment variable
; *LXA-ODF*baddir/LinkXall*
; *LXA-ODF*|SWDEV|/Libre/LinkXall/linkXall.odt*bm3  *Private environment variable
; *LXA-ODF*|SWDEV|Libre/LinkXall/linkXall.odt*bm4
; *LXA-ODF*** No recognized link because null uri
; *LXA-ODF*~/work/test/bm3.odt* Opens the file but bm is this comment
; *LXA-ODF*test/icky.odt*Bm30 * Bad file but possibly good bm. Test with/wo other file open.
; *LXA-ODF*../Libre/LinkXall/linkXall.odt*textOpenLink
; *LXA-ODF*../Libre/LinkXall/linkXall.odt** Placeholder for no bookmark

; .......................... TXT ...........................................
; *LXA-TXT*|SWDEV|Emacs/vMixed/linkxallSelf.el*#lxa-read-env
; *LXA-TXT*|SWDEV|Emacs/linkxall.el*#lxa-read-env
; *LXA-TXT*|UNDEFINED|hi.txt*
; *LXA-TXT*|SSH_AUTH_SOCK|/something.txt* Linux public environment
; *LXA-TXT*|LXADOC|link x all.el*End of progn wrapper

; .......................... BSR .......................................
; min and long forms. If DEFBROWSER is not defined the link is treated as WEB 
; with file protocol and, if Windows, any bookmark is dropped. WEB automatically
; moves the window to front but this happens with BSR only if the browser is 
; not already open. This is corrected using wintop.exe or wmctrl.
; *LXA-BSR*test/index.html*patents
; *LXA-BSR*|WEB|portfolio/index.html*patents
; *LXA-BSR*|LXADOC|/this was index.html*patents
; *LXA-BSR*|HOME|/work/test/index.html*experience*
; *LXA-BSR*~/work/test/index.html*experience
; *LXA-BSR*test/index.html*patents* Relative
; *LXA-BSR*c:/Web/portfolio/index.html*patents* Absolute Windows
; *LXA-BSR*/home/d/work/Web/portfolio/index.html*patents* Absolute Linux
; *LXA-BSR*unrealFile** Windows or Linux. Error reported by Emacs or Libre.
; *LXA-BSR*$HOME/work/test/index.html*patents* This does work in Windows.

; ........................... WEB ......................................
; *LXA-WEB*|HOME|/work/test/index.html*experience*
; *LXA-WEB*c:/home/work/test/index.html*patents* Semi-absolute
; *LXA-WEB*c:/home/work/test/index.html
; *LXA-WEB*~/work/test/index.html*experience
; *LXA-WEB*unrealFile** Windows or Linux. Error reported by Emacs or Libre.
; *LXA-WEB*file:///c:/Web/portfolio/index.html*experience* Windows abs with file protocol.
; *LXA-WEB*c:/Web/portfolio/index.html*experience* Windows abs w/o protocol.
; *LXA-WEB*file:///home/d/work/Web/portfolio/index.html*experience* Linux abs with protocol.
; *LXA-WEB*/home/d/work/Web/portfolio/index.html*experience* Linux abs without protocol.
; *LXA-WEB*file:///|HOME|/work/test/index.html** Env fails if uri with protocol.
; *LXA-WEB*|HOME|/work/test/index.html** Env ok if uri without protocol.
; *LXA-WEB*file:///test/index.html** Rel fails if uri with protocol.
; *LXA-WEB*test/index.html** Rel ok if uri w/o protocol.
; *LXA-WEB*file:///$HOME/work/test/index.html*patents* Semi-absolute with bm fails in Windows.
; *LXA-WEB*file:///~/work/test/index.html** Fails (with Emacs warning in Windows
;  but silently in Linux) because ~ is not expanded for LXA-WEB.
; *LXA-WEB*file:///home/d/work/Web/portfolio/index.html** Absolute Linux
; *LXA-WEB*http:///ixont.com/portfolio/index.html* * Opens the file in the website. 
; *LXA-WEB*http:///ixont.com/portfolio/index.html*patents* Opens the file in the website at the anchor. 
